#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['delver']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['colorful>=0.4.0', 'six>=1.10.0', 'terminaltables>=3.1.0']

entry_points = \
{'console_scripts': ['delve = delver.delve:main']}

setup(name='pydelver',
      version='0.0.6',
      description='Helpful tool for stepping through nested objects',
      author='Alex Sippel',
      author_email='asippel@narrativescience.com',
      url='https://github.com/NarrativeScience/delver',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6,<4',
     )
